import unittest
from typing import Sequence, Type

import equinox as eqx
import jax
import jax.numpy as jnp
import optax
import pytest
from jaxtyping import Array, Key

from thrml.block_management import Block
from thrml.block_sampling import SamplingSchedule, sample_states
from thrml.models.ising import (
    Edge,
    IsingEBM,
    IsingSamplingProgram,
    IsingTrainingSpec,
    estimate_kl_grad,
    hinton_init,
)
from thrml.pgm import AbstractNode, SpinNode


def get_double_grid(
    side_len: int,
    jumps: Sequence[int],
    n_visible: int,
    node: Type[AbstractNode],
    key: Key[Array, ""],
) -> tuple[Block, Block, Block, Block, list[AbstractNode], list[Edge]]:
    # n_groups = 2
    size = side_len**2
    assert n_visible <= size

    def get_idx(i, j):
        i = (i + side_len) % side_len
        j = (j + side_len) % side_len
        return i * side_len + j

    def get_coords(idx):
        return idx // side_len, idx % side_len

    def _make_edge(idx, di, dj):
        i, j = get_coords(idx)
        return jnp.array([idx, get_idx(i + di, j + dj)])

    make_edge = jax.jit(jax.vmap(_make_edge, in_axes=(0, None, None), out_axes=0))

    indices = jnp.arange(size)
    edges_arr = jnp.stack([indices, indices], axis=1)
    for d in jumps:
        left_edges = make_edge(indices, -d, 0)
        right_edges = make_edge(indices, d, 0)
        upper_edges = make_edge(indices, 0, -d)
        lower_edges = make_edge(indices, 0, d)
        edges_arr = jnp.concatenate([edges_arr, left_edges, right_edges, upper_edges, lower_edges], axis=0)

    deg = 4 * len(jumps) + 1
    total_edges = size * deg
    assert edges_arr.shape == (total_edges, 2)

    nodes_upper = [node() for _ in range(size)]
    nodes_lower = [node() for _ in range(size)]
    all_nodes = nodes_upper + nodes_lower
    all_edges = [(nodes_upper[i], nodes_lower[j]) for i, j in edges_arr]

    visible_indices = jax.random.permutation(key, jnp.arange(size))[:n_visible]
    visible_nodes = [nodes_upper[i] for i in visible_indices]
    upper_without_visible = [node for node in nodes_upper if node not in visible_nodes]

    return (
        Block(nodes_upper),
        Block(nodes_lower),
        Block(visible_nodes),
        Block(upper_without_visible),
        all_nodes,
        all_edges,
    )


@pytest.mark.slow
class TestTrainMnist(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_classes = [0, 3, 4]
        self.num_label_spots = 10
        label_size = len(self.target_classes) * self.num_label_spots
        data_dim = 28 * 28 + label_size

        self.train_data_filtered = jnp.load("tests/mnist_test_data/train_data_filtered.npy")
        self.sep_images_test = {}
        for digit in self.target_classes:
            self.sep_images_test[digit] = jnp.load(f"tests/mnist_test_data/sep_images_test_{digit}.npy")

        (upper_grid, lower_grid, visible_nodes, upper_without_visible, all_nodes, all_edges) = get_double_grid(
            40, [1, 4, 15], data_dim, SpinNode, jax.random.key(0)
        )

        self.init_model = IsingEBM(
            all_nodes,
            all_edges,
            jnp.zeros((len(all_nodes),), dtype=float),
            jnp.zeros((len(all_edges),), dtype=float),
            jnp.array(1.0),
        )

        self.positive_sampling_blocks = [upper_without_visible, lower_grid]
        self.negative_sampling_blocks = [upper_grid, lower_grid]
        self.training_data_blocks = [visible_nodes]

        image_block = Block(visible_nodes.nodes[: 28 * 28])
        upper_without_image = Block([node for node in upper_grid if node not in image_block.nodes])
        self.classification_sampling_blocks = [upper_without_image, lower_grid]
        self.classification_data_blocks = [image_block]
        self.classification_label_block = Block(visible_nodes.nodes[28 * 28 :])

        self.schedule_negative = SamplingSchedule(200, 40, 5)
        self.schedule_positive = SamplingSchedule(200, 20, 10)
        self.accuracy_schedule = SamplingSchedule(400, 40, 10)

        self.optim = optax.adam(learning_rate=0.01)
        self.n_epochs = 1

    def test_train_mnist(self):
        def do_epoch_simplified(
            key,
            model,
            bsz_positive,
            bsz_negative,
            data_positive,
            opt_state=None,
        ):
            def batch_data(key, data, _bsz, clamped_blocks):
                clamped_nodes = [node for block in clamped_blocks for node in block]
                data_size = data.shape[0]
                assert data.shape == (data_size, len(clamped_nodes))
                key, key_shuffle = jax.random.split(key)
                idxs = jax.random.permutation(key_shuffle, jnp.arange(data_size))
                data = data[idxs]
                _n_batches = data_size // _bsz
                tot_len = _n_batches * _bsz
                batched_data = jnp.reshape(data[:tot_len], (_n_batches, _bsz, len(clamped_nodes))).astype(jnp.bool)
                return batched_data, _n_batches

            key, key_pos = jax.random.split(key, 2)
            batched_data_pos, n_batches = batch_data(key_pos, data_positive, bsz_positive, self.training_data_blocks)

            def body_fun(carry, key_and_data):
                _key, _data_pos = key_and_data

                _opt_state, _params = carry
                _model = eqx.tree_at(lambda m: (m.weights, m.biases), model, _params)
                key_train, key_init_pos, key_init_neg = jax.random.split(_key, 3)
                vals_free_pos = hinton_init(key_init_pos, _model, self.positive_sampling_blocks, (1, bsz_positive))
                vals_free_neg = hinton_init(
                    key_init_neg,
                    _model,
                    self.negative_sampling_blocks,
                    (bsz_negative,),
                )

                ebm = IsingTrainingSpec(
                    _model,
                    self.training_data_blocks,
                    [],
                    self.positive_sampling_blocks,
                    self.negative_sampling_blocks,
                    self.schedule_positive,
                    self.schedule_negative,
                )

                grad_w, grad_b, _, _ = estimate_kl_grad(
                    key_train, ebm, _model.nodes, model.edges, [_data_pos], [], vals_free_pos, vals_free_neg
                )

                grads = (grad_w, grad_b)
                # optax does not obey this
                with jax.numpy_dtype_promotion("standard"):
                    updates, _opt_state = self.optim.update(grads, _opt_state, _params)
                _weights, _biases = _params
                # todo: grads are negative vs. positive
                _weights += updates[0]
                _biases += updates[1]

                new_carry = _opt_state, (_weights, _biases)

                return new_carry, None

            params = model.weights, model.biases

            init_carry = opt_state, params

            keys = jax.random.split(key, n_batches)
            out_carry, _ = jax.lax.scan(body_fun, init_carry, (keys, batched_data_pos))

            opt_state, params = out_carry
            new_model = eqx.tree_at(lambda m: (m.weights, m.biases), model, params)

            return new_model, opt_state

        def compute_accuracy(
            key,
            model,
            bsz_per_digit,
        ):
            """Takes images separated into classes based on which digit they are and
            for each class computes the probability that the model assigns a 1 to the label
            of each digit. Based on this it computes the accuracy (the fraction of
            digits where the argmax of the output labels gives the correct digit)
            and records the average label probabilities for each digit.
            """
            accuracy = 0.0
            for i, digit in enumerate(self.target_classes):
                images = self.sep_images_test[digit][:bsz_per_digit].astype(jnp.bool)

                program = IsingSamplingProgram(
                    model, self.classification_sampling_blocks, self.classification_data_blocks
                )

                key, subkey1, subkey2 = jax.random.split(key, 3)

                init_free_states = hinton_init(subkey2, model, self.classification_sampling_blocks, (bsz_per_digit,))

                keys_samp = jax.random.split(subkey1, bsz_per_digit)

                samples = jax.vmap(
                    lambda k, i, d: sample_states(
                        k, program, self.accuracy_schedule, i, d, [self.classification_label_block]
                    )
                )(keys_samp, init_free_states, [images])[0]

                labels = samples.reshape(
                    samples.shape[0],
                    samples.shape[1],
                    self.num_label_spots,
                    len(self.target_classes),
                )
                labels = jnp.mean(labels, axis=(1, 2))
                generated_digit = jnp.argmax(labels, axis=1)
                # note that the generated label will not be the digit itself,
                # but the index of the digit in the list of target digits
                accuracy += jnp.mean(i == generated_digit)

            accuracy /= len(self.target_classes)
            return accuracy

        best_accuracy = 0.0
        opt_state = self.optim.init((self.init_model.weights, self.init_model.biases))

        model = self.init_model

        for i in range(self.n_epochs):
            model, opt_state = do_epoch_simplified(
                jax.random.key(0),
                model,
                50,
                25,
                self.train_data_filtered,
                opt_state,
            )

            accuracy = compute_accuracy(
                jax.random.key(2),
                model,
                500,
            )
            best_accuracy = max(best_accuracy, accuracy)

        self.assertGreater(best_accuracy, 0.9)



import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import PyTree

from thrml.block_management import Block


class InteractionGroup(eqx.Module):
    """Defines computational dependencies for conditional sampling updates.

    An `InteractionGroup` specifies information that is required to update the state of some subset
    of the nodes of a PGM during a block sampling routine.

    More concretely, when the state of the node at head_nodes[i] is being updated, the sampler will receive the current
    state of the nodes at tail_nodes[k][i] for all k, and the ith element of each array in the Interaction PyTree
    (sliced along the first dimension).

    **Attributes:**

    - `head_nodes`: these are the nodes whose conditional updates should be affected by this InteractionGroup.
    - `tail_nodes`: these are the nodes whose state information is required to update `head_nodes`.
    - `interaction`: this specifies the parametric (independent of the state of the sampling program)
        required to update 'head_nodes'.
    """

    head_nodes: Block
    tail_nodes: list[Block]
    interaction: PyTree

    def __init__(self, interaction: PyTree, head_nodes: Block, tail_nodes: list[Block]):
        """Create an `InteractionGroup`.

        An `InteractionGroup` implements a group of directed interactions between nodes
        in a PGM sampling program.

        **Arguments:**

        - `interaction`: A PyTree specifying the static information associated with the
            interaction. The first dimension of every Array in interaction must be equal
            to the length of `head_nodes`.
        - `head_nodes`: The nodes whose update is affected by the interaction.
        - `tail_nodes`: The groups of nodes whose state is required to update
            `head_nodes`. Each block in this list of blocks is intended to be parallel
            to `head_nodes`. i.e, to update the state of head_nodes[i] during sampling
            we need state info about tail_nodes[k][i] for all values of k.
        """

        interaction_size = len(head_nodes.nodes)

        for block in tail_nodes:
            if not len(block.nodes) == interaction_size:
                raise RuntimeError("All tail node blocks must have the same length as head_nodes")

        def _get_dim(x):
            return (-1 if not len(x.shape) else x.shape[0]) if isinstance(x, jnp.ndarray) else interaction_size

        dims = jax.tree.leaves(jax.tree.map(_get_dim, interaction))
        if not all(dim == interaction_size for dim in dims):
            raise RuntimeError(
                "All arrays in interaction must have leading dimension equal to the length of head_nodes"
            )

        self.interaction = interaction
        self.head_nodes = head_nodes
        self.tail_nodes = tail_nodes



import unittest

import jax.numpy as jnp

from thrml.block_management import Block
from thrml.interaction import InteractionGroup
from thrml.pgm import AbstractNode


class Node(AbstractNode):
    pass


class TestInteractionInputs(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        block_size = 4
        self.good_head = Block([Node() for _ in range(block_size)])
        self.good_tails = [Block([Node() for _ in range(block_size)]) for _ in range(2)]
        self.good_interaction = jnp.zeros((block_size,))

    def test_good(self):
        _ = InteractionGroup(self.good_interaction, self.good_head, self.good_tails)

    def test_bad_tail(self):
        bad_tail = [Block([Node() for _ in range(len(self.good_head) + 1) for _ in range(2)]) for _ in range(2)]

        with self.assertRaises(RuntimeError) as error:
            _ = InteractionGroup(self.good_interaction, self.good_head, bad_tail)

        self.assertIn("tail node blocks", str(error.exception))

    def test_bad_interaction(self):
        bad_interaction = (self.good_interaction, jnp.array(1.0))
        with self.assertRaises(RuntimeError) as error:
            _ = InteractionGroup(bad_interaction, self.good_head, self.good_tails)

        self.assertIn("leading dimension", str(error.exception))

        bad_interaction = (self.good_interaction, jnp.zeros((len(self.good_head) + 1,)))
        with self.assertRaises(RuntimeError) as error:
            _ = InteractionGroup(bad_interaction, self.good_head, self.good_tails)
        self.assertIn("leading dimension", str(error.exception))



import unittest

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Key, PyTree

from thrml.block_management import Block
from thrml.block_sampling import (
    BlockGibbsSpec,
    BlockSamplingProgram,
    SamplingSchedule,
    sample_blocks,
    sample_single_block,
    sample_states,
)
from thrml.conditional_samplers import (
    AbstractConditionalSampler,
    _SamplerState,
    _State,
)
from thrml.interaction import InteractionGroup
from thrml.pgm import AbstractNode


class ContinousScalarNode(AbstractNode):
    pass


class PlusInteraction(eqx.Module):
    multiplier: Array


class MinusInteraction(eqx.Module):
    multiplier: Array


class MemoryInteraction(eqx.Module):
    multiplier: Array


class PlusMinusSampler(AbstractConditionalSampler):
    def sample(
        self,
        key,
        interactions: list[PyTree],
        active_flags: list[Array],
        states: list[list[_State]],
        sampler_state: _SamplerState,
        output_sd: jax.ShapeDtypeStruct,
    ):
        output = jnp.zeros(output_sd.shape, dtype=output_sd.dtype)
        for interaction, active, state in zip(interactions, active_flags, states):
            active = active.astype(interaction.multiplier.dtype)
            s = state[0].astype(interaction.multiplier.dtype)
            if isinstance(interaction, (PlusInteraction, MemoryInteraction)):
                output += jnp.sum(interaction.multiplier * active * s, axis=-1)
            elif isinstance(interaction, MinusInteraction):
                output -= jnp.sum(interaction.multiplier * active * s, axis=-1)
            else:
                raise RuntimeError("Invalid interaction passed to PlusMinusSampler")

        return output, sampler_state

    def init(self) -> _SamplerState:
        return None


class TestPlusMinus(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        key = jax.random.key(424)

        free_nodes = [ContinousScalarNode() for _ in range(3)]

        minus_nodes = [ContinousScalarNode() for _ in range(2)]
        plus_nodes = [ContinousScalarNode() for _ in range(2)]

        key, subkey = jax.random.split(key, 2)
        self.minus_weights = jax.random.uniform(subkey, (3,), minval=0, maxval=1)
        key, subkey = jax.random.split(key, 2)
        self.plus_weights = jax.random.uniform(subkey, (3,), minval=0, maxval=1)

        minus_interaction_group = InteractionGroup(
            MinusInteraction(self.minus_weights),
            Block([free_nodes[0], free_nodes[0], free_nodes[1]]),
            [Block([minus_nodes[0], minus_nodes[1], minus_nodes[1]])],
        )

        plus_interaction_group = InteractionGroup(
            PlusInteraction(self.plus_weights),
            Block([free_nodes[1], free_nodes[2], free_nodes[2]]),
            [Block([plus_nodes[0], plus_nodes[0], plus_nodes[1]])],
        )

        memory_interaction_group = InteractionGroup(
            MemoryInteraction(jnp.ones(len(free_nodes))), Block(free_nodes), [Block(free_nodes)]
        )

        block_spec = BlockGibbsSpec(
            [Block([free_nodes[0]]), Block(free_nodes[1:])],
            [Block(plus_nodes + minus_nodes)],
            {ContinousScalarNode: jax.ShapeDtypeStruct((), jnp.float32)},
        )

        self.program = BlockSamplingProgram(
            block_spec,
            [PlusMinusSampler(), PlusMinusSampler()],
            [minus_interaction_group, plus_interaction_group, memory_interaction_group],
        )

        keys = jax.random.split(key, 4)

        self.state_free = [
            jax.random.uniform(keys[0], (1,), minval=1.0, maxval=5.0),
            jax.random.uniform(keys[1], (2,), minval=1.0, maxval=5.0),
        ]
        self.state_clamped = [jax.random.uniform(keys[2], (4,), minval=1.0, maxval=5.0)]

        self.key = keys[-1]

    def test_sample_block(self):
        outputs = []
        for block in [0, 1]:
            outputs.append(
                sample_single_block(self.key, self.state_free, self.state_clamped, self.program, block, None)[0]
            )

        first_output = self.state_free[0][0] - jnp.sum(self.minus_weights[:2] * self.state_clamped[0][2:])
        second_output = (
            self.state_free[1][0]
            - self.minus_weights[-1] * self.state_clamped[0][-1]
            + self.plus_weights[0] * self.state_clamped[0][0]
        )
        third_output = self.state_free[1][1] + jnp.sum(self.plus_weights[1:] * self.state_clamped[0][:2])

        self.assertTrue(np.allclose(outputs[0], [first_output], rtol=1e-6))
        self.assertTrue(np.allclose(outputs[1], [second_output, third_output], rtol=1e-6))

    def test_sample_blocks(self):
        # just make sure this runs, nothing new to learn
        sample_blocks(self.key, self.state_free, self.state_clamped, self.program, [None, None])

    def test_sample_states(self):
        # just make sure this runs
        schedule = SamplingSchedule(5, 5, 5)
        sample_states(
            self.key, self.program, schedule, self.state_free, self.state_clamped, self.program.gibbs_spec.free_blocks
        )

    def test_state_gaurdrailing(self):
        wrong_state_free = [self.state_free[0], jnp.zeros((2,), dtype=jnp.bool)]

        wrong_state_clamped = [jnp.zeros((4,), dtype=jnp.bool)]

        with self.assertRaises(RuntimeError) as error:
            _ = sample_blocks(self.key, wrong_state_free, self.state_clamped, self.program, [None, None])

        self.assertIn("type", str(error.exception))

        with self.assertRaises(RuntimeError) as error:
            _ = sample_blocks(self.key, self.state_free, wrong_state_clamped, self.program, [None, None])

        self.assertIn("type", str(error.exception))


class MultiNode(AbstractNode):
    pass


class MultiNodeState(eqx.Module):
    float_counter: Array
    cat_counter: Array


class IncrementSampler(AbstractConditionalSampler):
    def sample(
        self,
        key: Key,
        interactions: list[PyTree],
        active_flags: list[Array],
        states: list[list[_State]],
        sampler_state: _SamplerState,
        output_sd: PyTree[jax.ShapeDtypeStruct],
    ):
        assert isinstance(output_sd, MultiNodeState)

        for interaction, active, state in zip(interactions, active_flags, states):
            if isinstance(interaction, PlusInteraction):
                return (
                    MultiNodeState(state[0].float_counter[:, 0, :] + 1, state[0].cat_counter[:, 0, :] + 1),
                    sampler_state,
                )

    def init(self) -> _SamplerState:
        return None


class TestPyTreeState(unittest.TestCase):
    """Test that a sampling program involving nodes with more general pytree shapedtypes works"""

    def test_pytree_state(self):
        n_float = 2
        n_cat = 4

        sd_map = {
            MultiNode: MultiNodeState(
                jax.ShapeDtypeStruct((n_float,), jnp.float32), jax.ShapeDtypeStruct((n_cat,), jnp.int8)
            )
        }

        nodes = [MultiNode() for _ in range(10)]

        key = jax.random.key(424)

        interaction_group = InteractionGroup(PlusInteraction(jnp.ones((len(nodes),))), Block(nodes), [Block(nodes)])

        spec = BlockGibbsSpec([Block(nodes)], [], sd_map)

        key, subkey = jax.random.split(key, 2)
        init_float = jax.random.normal(subkey, (len(nodes), n_float))
        key, subkey = jax.random.split(key, 2)
        init_cat = jax.random.randint(subkey, (len(nodes), n_cat), minval=-4, maxval=4)

        init_state = [MultiNodeState(init_float, init_cat)]

        prog = BlockSamplingProgram(spec, [IncrementSampler()], [interaction_group])

        res, _ = sample_single_block(key, init_state, [], prog, 0, None)

        self.assertTrue(jnp.allclose(init_state[0].cat_counter + 1, res.cat_counter))
        self.assertTrue(jnp.allclose(init_state[0].float_counter + 1, res.float_counter))
